'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all libraries
import numpy as np
import pandas as pd
from scipy.stats import norm
import sys

from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class MTDF:
    def __init__(self, binary_columns : list = None, seed : 'int > 0' = 42) -> None:
        '''
            The constructor of the MTDF.
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        :param seed: int > 0, default = 42
            The seed used for random number generator.
        '''
        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns
        self.seed = seed
        np.random.seed(self.seed)

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''
        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''
        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def balance(self, df : pd.DataFrame, target : str):
        '''
            The balance function.
        :param df: pd.DataFrame
            The pandas Data Frame to apply the balancer.
        :param target: str
            The name of the target column.
        :return: pd.DataFrame
            A pandas Data Frame
        '''
        # Creating an internal copy of the data frame.
        self.df = df.copy()
        self.target = target

        # Checking if the target string based t algorithm is present in the data frame.
        if target not in self.df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        # Checking if the target column is a binary one.
        if len(self.df[target].unique()) != 2:
            raise NotBinaryData(f"{target} column isn't a binary column")

        # Getting the column names that are not the target one.
        self.X_columns = [column for column in self.df.columns if column != target]

        # Calculating the variance vector.
        variance = self.df[self.X_columns].var(axis=0).values

        # Calculating the Uset.
        uset = (self.df[self.X_columns].min(axis=0).values + self.df[self.X_columns].max(axis=0).values) / 2

        # Getting the classes frequency.
        classes_frequency = dict(self.df[target].value_counts())

        # Searching for the class with the biggest frequency.
        max_freq = 0
        for cls in classes_frequency:
            if classes_frequency[cls] > max_freq:
                majority_class = cls
                max_freq = classes_frequency[cls]

        # Getting the name of the minority class.
        minority_class = [cls for cls in classes_frequency if cls != majority_class][0]

        # Computing the lower and upper skewness.
        skewL = len(self.df[df[self.target] == minority_class]) / len(df)
        skewU = len(self.df[df[self.target] == majority_class]) / len(df)

        # Computing the lower and upper limits.
        a = self.df[self.X_columns].min(axis=0).values / 10
        b = self.df[self.X_columns].max(axis=0).values * 10

        # Updating the lower and upper limits.
        a = uset - skewL * np.sqrt(-2 * (variance / len(self.df[self.df[self.target] == minority_class])) * np.log(norm.cdf(a)))
        b = uset - skewU * np.sqrt(-2 * (variance / len(self.df[self.df[self.target] == majority_class])) * np.log(norm.cdf(b)))

        # Calculating the number of elements to generate.
        number_of_elements_to_generate = len(self.df[self.df[self.target] == majority_class]) - len(self.df[self.df[self.target] == minority_class])

        # Generating the synthetic data.
        self.synthetic_data = []

        # Eliminating possible -infinity and +infinity values.
        a = self.__infinity_check(a)
        b = self.__infinity_check(b)

        for _ in range(number_of_elements_to_generate):
            self.synthetic_data.append(
                # Generating arrays with random values between a and b.
                np.array([
                    np.random.uniform(a[i], b[i], 1)[0] for i in range(len(a))
                ])
            )

        # Replacing infinity values with minimal and maximal float python values.
        self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data).astype(float))

        # Creating the synthetic data frame.
        self.synthetic_df = pd.DataFrame(self.synthetic_data, columns=self.X_columns)

        # Rounding binary columns if needed.
        if self.__binarize:
            self.__to_binary()

        # Adding the target column to the synthetic data frame.
        self.synthetic_df.loc[:, self.target] = minority_class

        # Concatting the real data frame with the synthetic one.
        new_df = pd.concat([self.df, self.synthetic_df], axis=0)
        return new_df


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all libraries
import numpy as np
import pandas as pd
from functools import reduce
from operator import or_
from numba import njit
import random
import sys

from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

@njit
def index(array, item):
    for idx, val in np.ndenumerate(array):
        if val == item:
            return idx

class MWMOTE:
    def __init__(self, binary_columns : list = None, k1 : 'int > 1' = 5, k2 : 'int > 1' = 5, k3 : 'int > 1' = 5, Cth : 'int > 0' = 5, CMAX : 'int > 0' = 2, M : 'int' = 5, seed : 'int > 0' = 42):
        '''
            The constructor of the MWMOTE.
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        :param k1: int > 1, default = 5
            Number of neighbors used for predicting noisy minority class samples.
        :param k2: int > 1, default = 5
            Number of majority neighbors used for constructing informative minority set.
        :param k3: int > 1, default = 5
            Number of minority neighbors used for constructing informative minority set.
        :param Cth: int > 0, default = 5
            The threshold value of the closeness factor.
        :param CMAX: int > 0, default = 5
            Used in smoothing and re-scaling the values of different scaling factors.
        :param M: int, default = 5
            The number of clusters.
        :param seed: int > 0, default = 42
            The seed for the random number generator.
        '''
        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns
        self.__k1 = k1
        self.__k2 = k2
        self.__k3 = k3
        self.__Cth = Cth
        self.__CMAX = CMAX
        self.__M = M
        self.__seed = seed
        random.seed(self.__seed)

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''
        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''
        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def __predict_knn(self, sample, k, cls = 'auto'):
        '''
            The knn algorithm in one function.
        :param sample: 1-array
            The array for which we should find the nearest neighbours.
        :param k: int
            The number of neighbours.
        :param cls: str, default = 'auto'
            The class that we are interested in:
                If set to 'auto' then all classes are taken
                Else is taken the class passed.
        :return: array
            The indexes of the samples that are the nearest to the sample.
        '''
        if cls == 'auto':
            df = self.df[self.X_columns].values
        else:
            df = self.df[self.df[self.target] == cls][self.X_columns].values
        distances = []
        for x in df:
            distances.append(np.linalg.norm(x - sample, ord=2))
        predicted_index = np.argsort(distances)[1:k + 1]
        return predicted_index

    def __closeness_factor(self, y, x):
        '''
            The closeness factor function.
        :param y:
            The target class.
        :param x:
            The sample.
        :return:
            The closeness factor.
        '''
        dist = np.linalg.norm(y - x, ord=2)
        dist = dist / len(y)
        f = lambda x : x if x <= self.__Cth else self.__Cth
        return f(1 / dist) * self.__CMAX / self.__Cth

    def __density_factor(self, y, x):
        '''
            The density factor function.
        :param y:
            The target class.
        :param x:
            The sample.
        :return:
            The density factor.
        '''
        return self.__closeness_factor(y, x) / np.sum([self.__closeness_factor(y, q) for q in self.informative_minority_set])

    def __information_weight(self, y, x):
        '''
            The information weight function.
        :param y:
            The target class.
        :param x:
            The sample.
        :return:
            The information weight.
        '''
        return self.__closeness_factor(y, x) * self.__density_factor(y, x)

    def __dbscan(self, minority_class_sample):
        '''
            The BDSCAN fit function.
        :param minority_class_sample:
            The set of minority class samples
        :return: list
            The list of clusters of every sample from the set.
        '''
        clusters = np.arange(len(minority_class_sample))
        while len(set(clusters)) > self.__M:
            first = np.random.randint(0, len(clusters))
            second = np.random.randint(0, len(clusters))
            if first == second:
                continue
            min_distance = np.linalg.norm(minority_class_sample[first] - minority_class_sample[second], ord=2)
            closest_pair = [first, second]
            for i in range(len(minority_class_sample)):
                for j in range(len(minority_class_sample)):
                    if clusters[i] == clusters[j]:
                        continue
                    elif min_distance >= np.linalg.norm(minority_class_sample[i] - minority_class_sample[j], ord=2) and np.linalg.norm(minority_class_sample[i] - minority_class_sample[j], ord=2) != 0:
                        min_distance = np.linalg.norm(minority_class_sample[i] - minority_class_sample[j], ord=2)
                        closest_pair = [clusters[i], clusters[j]]
            to_replace = closest_pair[0]
            clusters[clusters == to_replace] = closest_pair[1]
        return clusters

    def __pridict_dbscan(self, sample, minority_class_sample, clusters):
        '''
            The DBSCAN predict function.
        :param sample: array
            The array which we must cluster.
        :param minority_class_sample: 2-d array
            The minority set.
        :param clusters: array
            The array with clusters for every sample.
        :return: int
            The cluster of the sample.
        '''
        dist = []
        for x in minority_class_sample:
            dist.append(np.linalg.norm(x - sample, ord=2))
        return clusters[np.argsort(np.array(dist))[0]]

    def balance(self, df : pd.DataFrame, target : str):
        '''
            The balance function.
        :param df: pd.DataFrame
            The pandas Data Frame to apply the balancer.
        :param target: str
            The name of the target column.
        :return: pd.DataFrame
            A pandas Data Frame
        '''

        # Creating an internal copy of the data frame.
        self.df = df.copy()
        self.target = target

        # Checking if the target string based t algorithm is present in the data frame.
        if target not in self.df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        # Checking if the target column is a binary one.
        if len(self.df[target].unique()) != 2:
            raise NotBinaryData(f"{target} column isn't a binary column")

        # Getting the column names that are not the target one.
        self.X_columns = [column for column in self.df.columns if column != target]

        # Getting the class frequencies.
        classes_frequency = dict(self.df[target].value_counts())

        # Searching for the class with the biggest frequency.
        max_freq = 0
        for cls in classes_frequency:
            if classes_frequency[cls] > max_freq:
                majority_class = cls
                max_freq = classes_frequency[cls]

        # Getting the name of the minority class.
        minority_class = [cls for cls in classes_frequency if cls != majority_class][0]

        # Getting the data for the minority class set.
        minority_set = self.df[self.df[self.target] == minority_class][self.X_columns].values

        # Defining the empty filtered minority set.
        filtered_minority_set = []

        # Constructing the filtered minority set
        for i in range(len(minority_set)):
            neightbours_index = self.__predict_knn(minority_set[i], self.__k1)
            if minority_class not in dict(self.df.iloc[neightbours_index, :][self.target].value_counts()).values():
                filtered_minority_set.append(i)

        # Defining the empty nearest neighbours list.
        NearestNeighbours = []

        # Searching for the k2 nearest neighbours for every sample in filtered minority set
        for i in filtered_minority_set:
            NearestNeighbours.append(self.__predict_knn(self.df.iloc[i, :][self.X_columns].values, self.__k2, majority_class))

        # Converting the NearestNeighbours into a list of sets.
        NearestNeighbours = [set(element.tolist()) for element in NearestNeighbours]

        # Getting the border line majority set
        border_line_majority_set = list(reduce(or_, NearestNeighbours))

        # Clearing the NearestNeighbours list
        NearestNeighbours = []

        # Getting the k3 nearest neighbours for every sample from border line majority set
        for x in border_line_majority_set:
            NearestNeighbours.append(self.__predict_knn(x, self.__k3, minority_class))

        # Converting the NearestNeighbours into a list of sets.
        NearestNeighbours = [set(element.tolist()) for element in NearestNeighbours]

        # Getting the informative minority set.
        self.informative_minority_set = np.array(list(reduce(or_, NearestNeighbours)))

        # Getting the selection weights.
        selection_weights = np.array([np.sum([self.__information_weight(self.df[self.df[self.target] == majority_class].values[y],
                                                                        self.df[self.df[self.target] == minority_class].values[x])
                                              for y in border_line_majority_set]) for x in self.informative_minority_set])
        # Getting the selection probabilities.
        selection_probability = selection_weights / np.sum(selection_weights)

        # Getting the clusters from dbscan.
        clusters = self.__dbscan(minority_set)

        # Defining the empty list with synthetic data.
        self.synthetic_data = []

        # Getting the number of new samples to generate.
        N = len(self.df[self.df[self.target] == majority_class]) - len(self.df[self.df[self.target] == minority_class])

        # Generating the new samples.
        for i in range(N):
            # Getting the first random sample from minority set.
            first = self.df.iloc[np.random.choice(self.informative_minority_set, 1, p = selection_probability), :][self.X_columns].values[0]

            # Getting the cluster of the first selected sample.
            cluster = self.__pridict_dbscan(first, minority_set, clusters)

            if len(clusters[clusters == cluster]) >= 2:
                # Getting the second random sample from the same cluster.
                second = self.df.iloc[np.random.choice(
                    self.informative_minority_set[
                        list(index(clusters, cluster))], 1),
                         :][self.X_columns].values[0]
            else:
                second = self.df.iloc[self.__predict_knn(first, self.__k1, minority_class), :][self.X_columns].values[0]

            # Appending the generated point to the synthetic data.
            self.synthetic_data.append(first + random.uniform(0, 1) * (second - first))

        # Replacing infinity values with minimal and maximal float python values.
        self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data).astype(float))

        # Creating the synthetic data frame.
        self.synthetic_df = pd.DataFrame(np.array(self.synthetic_data), columns=self.X_columns)

        # Rounding binary columns if needed.
        if self.__binarize:
            self.__to_binary()

        # Adding the target column.
        self.synthetic_df.loc[:, self.target] = minority_class
        new_df = pd.concat([self.df, self.synthetic_df], axis=0)
        return new_df


'''
Created with love by Sigmoid

@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
import random
import sys
from random import randrange
from .SMOTE import SMOTE
from sklearn.mixture import GaussianMixture
from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn


class SCUT:


    def __init__(self,k: "int > 0" = 5, seed: float = 42, binary_columns : list = None) -> None:
        '''
            Setting up the algorithm
        :param k: int, k>0, default = 5
            Number of neighbours which will be considered when looking for simmilar data points
        :param seed: intt, default = 42
            seed for random
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        '''

        self.__k = k

        if binary_columns is None:
            self.__binarize = False
            self.__binary_columns = None
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns

        self.__seed = seed
        np.random.seed(self.__seed)
        random.seed(self.__seed)

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''

        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''

        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def balance(self, df : pd.DataFrame, target : str):
        '''
            Reducing the dimensionality of the data
        :param df: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The target name of the value that we have to predict
        '''

        #get unique values from target column
        unique = df[target].unique()

        if target not in df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        self.target= target
        self.df = df.copy()

        #training columns
        self.X_columns = [column for column in self.df.columns if column != target]

        class_samples = []
        for clas in unique:
            class_samples.append(self.df[self.df[self.target]==clas][self.X_columns].values)


        classes_nr_samples = []
        for clas in unique:
            classes_nr_samples.append(len(self.df[self.df[self.target]==clas]))

        #getting mean number of samples of all classes
        mean = np.mean(classes_nr_samples)

        #undersampling by SCUT algorithm
        for i,class_sample in enumerate(class_samples):

            #cheching for
            if classes_nr_samples[i]>mean:
                clusters = []
                selected_samples = []

                #getting nr of samples to take
                difference = classes_nr_samples[i] - mean

                #clusster every class
                gmm = GaussianMixture(3)
                gmm.fit(class_sample)

                #getting predictions
                labels = gmm.predict(class_sample)
                clusters_df = pd.DataFrame(np.array(class_sample), columns=self.X_columns)
                clusters_df.loc[:, target] = labels 
                unique_clusters = clusters_df[self.target].unique()

                #Selecting random samples from every cluster, and repeat untill we get the "difference" that we needed
                for clas in unique_clusters:

                    #clusster sets
                    clusters.append(clusters_df[clusters_df[self.target]==clas].values)

                while difference != 0:
                    for cluster in clusters:

                        #selecting random sample and add it to list
                        index = randrange(len(cluster))
                        example = cluster[index]
                        selected_samples.append(example)
                        difference-=1
                        if difference == 0:
                            break

                #create the new dataset including selected samples
                cluster_df = pd.DataFrame(np.array(selected_samples), columns=self.df.columns)
                cluster_df.loc[:, self.target] = unique[i]##
                if i==0:
                    self.new_df = cluster_df
                else:
                    self.new_df = pd.concat([cluster_df,self.new_df],axis=0)
            else:
                
                cluster_df = pd.DataFrame(np.array(class_sample), columns=self.X_columns)
                cluster_df.loc[:, self.target] = unique[i]##
                if i==0:
                    self.new_df = cluster_df
                else:
                    self.new_df = pd.concat([cluster_df,self.new_df],axis=0)


        #oversampling using SMOTE
        smot = SMOTE(binary_columns = self.__binary_columns)

        #getting finall dataset
        self.new_df = smot.balance(self.new_df,self.target)

        #return new dataset
        return self.new_df


# Importing all balancers from modules.
from .ADASYN import ADASYN
from .ICOTE import ICOTE
from .MTDF import MTDF
from .MWMOTE import MWMOTE, index
from .SCUT import SCUT
from .SLS import SLS
from .SMOTE import SMOTE
from .SMOTEENN import SMOTEENN
from .SMOTENC import SMOTENC
from .SMOTETOMEK import SMOTETOMEK
from .TKRKNN import TKRKNN



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all libraries
import numpy as np
import pandas as pd
import sys
from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class TKRKNN:
    def __init__(self, binary_columns : list = None, seed : 'int > 0' = 42, k : 'int > 0' = 5) -> None:
        '''
            The constructor of TRKNN.
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        :param seed: int > 0, default = 42
            The seed for the random number generator.
        :param k: int > 0, default = 5
            The number of neighbours used by the knn algorithm.
        '''
        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns
        self.__seed = seed
        self.__k = k
        np.random.seed(self.__seed)
    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''
        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''
        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def __predict_knn(self, sample, cls):
        '''
            The knn algorithm in one function.
        :param sample: 1-array
            The array for which we should find the nearest neighbours.
        :param cls: object
            The class that we want to query.
        :return: array
            The indexes of the samples that are the nearest to the sample.
        '''
        distances = []
        for x in self.df[self.df[self.target] == cls][self.X_columns].values:
            distances.append(np.linalg.norm(x - sample, ord=2))
        predicted_index = np.argsort(distances)[1:self.__k + 1]
        return predicted_index

    def balance(self, df : 'pd.DataFrame', target : str):
        '''
            The balance function.
        :param df: pd.DataFrame
            The pandas Data Frame to apply the balancer.
        :param target: str
            The name of the target column.
        :return: pd.DataFrame
            A pandas Data Frame
        '''

        # Creating an internal copy of the data frame.
        self.df = df.copy()
        self.target = target

        # Checking if the target string based t algorithm is present in the data frame.
        if target not in self.df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        # Checking if the target column is a binary one.
        if len(self.df[target].unique()) != 2:
            raise NotBinaryData(f"{target} column isn't a binary column")

        # Getting the column names that are not the target one.
        self.X_columns = [column for column in self.df.columns if column != target]

        # Getting the class frequencies.
        classes_frequency = dict(self.df[target].value_counts())

        # Searching for the class with the biggest frequency.
        max_freq = 0
        for cls in classes_frequency:
            if classes_frequency[cls] > max_freq:
                majority_class = cls
                max_freq = classes_frequency[cls]

        # Getting the name of the minority class.
        minority_class = [cls for cls in classes_frequency if cls != majority_class][0]

        # Getting the sample set of the minority class.
        minority_set = self.df[self.df[self.target] == minority_class][self.X_columns].values

        # Defining the empty nearest neighbours array
        nnarray = []

        # Searching for k nearest neighbours for every sample in the minority class.
        for x in minority_set:
            nnarray.append(self.__predict_knn(x, minority_class))

        # Computing the number of samples to generate.
        samples_to_generate = len(self.df[self.df[self.target] == majority_class]) - len(self.df[self.df[self.target] == minority_class])

        # Defining the empty list for synthetic data.
        self.synthetic_data = []

        # Generating of the synthetic data.
        while samples_to_generate != 0:
            # Randomly selecting a sample from minority set.
            nn = np.random.randint(low=0, high=self.__k, size=1)
            random_index = np.random.randint(low=0, high=len(minority_set), size=1)[0]
            self.synthetic_data.append(minority_set[random_index])

            # Randomly change the attributes of the selected sample.
            for attr_index in range(len(self.X_columns)):
                dif = minority_set[nnarray[random_index][nn], attr_index] - minority_set[random_index, attr_index]
                gap = np.random.uniform(size=1)
                self.synthetic_data[-1][attr_index] = minority_set[random_index, attr_index] + gap * dif
            samples_to_generate -= 1

        # Replacing infinity values with minimal and maximal float python values.
        self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data).astype(float))

        # Creating the synthetic data frame.
        self.synthetic_df = pd.DataFrame(self.synthetic_data, columns=self.X_columns)

        # Rounding binary columns if needed.
        if self.__binarize:
            self.__to_binary()

        # Adding the target column.
        self.synthetic_df.loc[:, self.target] = minority_class
        new_df = pd.concat([self.df, self.synthetic_df], axis=0)
        return new_df


'''
Created with love by Sigmoid

@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
import random
import sys
from random import randrange
from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass

import warnings
warnings.warn = warn


class SMOTETOMEK:


    def __init__(self,k: "int > 0" = 5, seed: float = 42, binary_columns : list = None) -> None:
        '''
            Setting up the algorithm
        :param k: int, k>0, default = 5
            Number of neighbours which will be considered when looking for simmilar data points
        :param seed: intt, default = 42
            seed for random
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        '''

        self.__k = k

        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns

        self.__seed = seed
        np.random.seed(self.__seed)
        random.seed(self.__seed)

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''

        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''

        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def __undersample(self):
        '''
            Undersample the dataset using Tomek links algorithm
        '''

        distances = []
        minority_nearest = []
        majority_nearest = []

        #save distances for minority class
        for example in self.minority_samples:
            distances = []
            for x in self.majority_samples:
                distances.append(np.linalg.norm(x - example, ord=2))
            predicted_index = np.argsort(distances)[0]
            minority_nearest.append(predicted_index)

        #save distances for majority class
        for example in self.majority_samples:
            distances = []
            for x in self.minority_samples:
                distances.append(np.linalg.norm(x - example, ord=2))
            predicted_index = np.argsort(distances)[0]
            majority_nearest.append(predicted_index)

        #find all tomek links
        to_delete = []
        for i in range(len(minority_nearest)):
            if i == majority_nearest[minority_nearest[i]]:
                to_delete.append(minority_nearest[i])

        #delete all Tomek links
        self.majority_samples = np.delete(self.majority_samples,to_delete,axis = 0)

        # Creating new dataset without Tomek Links
        mj = pd.DataFrame(self.majority_samples,columns = self.X_columns)
        mj.loc[:, self.target] = self.majority_class
        mn = pd.DataFrame(self.minority_samples,columns = self.X_columns)
        mn.loc[:, self.target] = self.minority_class
        self.df = pd.concat([mj,mn],axis = 0)


    def balance(self, df : pd.DataFrame, target : str)-> pd.DataFrame:
        '''
            Reducing the dimensionality of the data
        :param df: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The target name of the value that we have to predict
        '''

        #check for binary
        unique = df[target].unique()
        if len(unique)!=2:
            raise NotBinaryData(f"{target} column isn't a binary column")

        if target not in df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        self.target= target
        self.df = df.copy()
        self.X_columns = [column for column in self.df.columns if column != target]

        #check the minority class
        first_class = len(df[df[target]==unique[0]])/len(df[target])
        if first_class > 0.5:
            self.minority_class,self.majority_class = unique[1],unique[0]
        else: 
            self.minority_class,self.majority_class = unique[0],unique[1]

        majority = df[df[target]==self.majority_class]
        minority = df[df[target]==self.minority_class]

        self.minority_samples = self.df[self.df[target] == self.minority_class][self.X_columns].values
        self.majority_samples = self.df[self.df[target] == self.majority_class][self.X_columns].values

        #calling undersample function
        self.__undersample()

        #find difference 
        difference = len(majority)-len(minority)
        self.synthetic_data = []
        for _ in range(difference):

            #random example from minority class
            index = randrange(len(self.minority_samples))
            example = self.minority_samples[index]

            #select k neighbouors from this example
            neighbours_indexes = self.__get_k_neighbours(example) 

            #select random neighbour
            index = randrange(len(neighbours_indexes))
            selected_neighbour = neighbours_indexes[index]
            selected_neighbour = self.minority_samples[selected_neighbour]

            #select random point between example and neighbour
            alpha = random.random()
            new_example = example+alpha*(selected_neighbour-example)

            #add it to df
            self.synthetic_data.append(new_example)

        self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data))
        self.synthetic_df = pd.DataFrame(np.array(self.synthetic_data), columns=self.X_columns)
        self.synthetic_df.loc[:, target] = self.minority_class
        self.synthetic_df = pd.concat([self.synthetic_df,self.df],axis=0)

        # Rounding binary columns if needed.
        if self.__binarize:
            self.__to_binary()

        #return new df
        return self.synthetic_df
    
    def __get_k_neighbours(self,example):
        '''
            KNN, getting nearest neighbors
        :param example: Numpy array
            the sample row from minority class to get neighbours from
        :param minority_samples: Numpy.ndarray
            minority class samples from where we find neighbours
        '''
        distances = []
        
        for x in self.minority_samples:
            distances.append(np.linalg.norm(x - example, ord=2))
        predicted_index = np.argsort(distances)[1:self.__k + 1]

        return predicted_index


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
class NotBetweenZeroAndOneError(BaseException):
    ''' Raised when a value is not between zero and one, but it should be'''
    pass

class NotBinaryData(BaseException):
    ''' Raised when the data passed is not binary '''
    pass

class NoSuchMethodError(BaseException):
    ''' Raised when the balancer or reducer doesnt't have a method '''
    pass

class NoSuchColumn(BaseException):
    ''' Raised when the data frame passed doesn't have a certain column '''
    pass


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all libraries
import numpy as np
import pandas as pd
from math import floor
import sys
from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class ICOTE:
    def __init__(self, binary_columns : list = None, seed : 'int > 0' = 42) -> None:
        '''
            The constructor of ICOTE.
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        :param seed: int > 0, default = 42
            The seed for random number generator.
        '''
        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns
        self.__seed = seed
        np.random.seed(self.__seed)

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''
        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''
        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def balance(self, df : 'pd.DataFrame', target : str):
        '''
            The balance function.
        :param df: pd.DataFrame
            The pandas Data Frame to apply the balancer.
        :param target: str
            The name of the target column.
        :return: pd.DataFrame
            A pandas Data Frame
        '''

        # Creating an internal copy of the data frame.
        self.df = df.copy()
        self.target = target

        # Checking if the target string based t algorithm is present in the data frame.
        if target not in self.df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        # Checking if the target column is a binary one.
        if len(self.df[target].unique()) != 2:
            raise NotBinaryData(f"{target} column isn't a binary column")

        # Getting the column names that are not the target one.
        self.X_columns = [column for column in self.df.columns if column != target]

        # Getting the class frequencies.
        classes_frequency = dict(self.df[target].value_counts())

        # Searching for the class with the biggest frequency.
        max_freq = 0
        for cls in classes_frequency:
            if classes_frequency[cls] > max_freq:
                majority_class = cls
                max_freq = classes_frequency[cls]

        # Getting the name of the minority class.
        minority_class = [cls for cls in classes_frequency if cls != majority_class][0]

        # Splitting the data in features and target arrays.
        X, y = self.df[self.X_columns].values, self.df[self.target].values

        # Getting the min and max arrays of the features.
        X_min = X.min(axis=0)
        X_max = X.max(axis=0)

        # Normalising the data.
        norm_X = (X - X_min) / (X_max - X_min)

        # Getting the indexes of the minority and majority classes.
        minority_index = np.where(y == minority_class)
        majority_index = np.where(y == majority_class)

        # Getting the number of samples of minority class.
        minority_samples: int = len(minority_index)

        # Defining the empty list with synthetic data.
        self.synthetic_data = []
        while minority_samples <= len(majority_index):
            # Creating clones for every minority sample.
            for i in minority_index[0]:
                clones = [].copy()
                dist = [].copy()
                for j in majority_index[0]:
                    dist.append(np.linalg.norm(norm_X[i] - norm_X[j], ord=2))
                for dist_index in range(len(dist)):
                    if dist[dist_index] == 0:
                        continue
                    for _ in range(floor(dist[dist_index])):
                        clones.append([norm_X[i], dist_index])
                for clone_index in range(len(clones)):
                    alpha = 1 / dist[clones[clone_index][1]]
                    if alpha < 0.05:
                        continue
                    # Creating mutations in clones.
                    clones[clone_index][0] = clones[clone_index][0] + alpha * (clones[clone_index][0] - norm_X[clones[clone_index][1]])
                # Adding the clones to the synthetic data list.
                self.synthetic_data.extend([clone[0] for clone in clones])
                minority_samples += len(clones)
        # Selecting majority - minority samples.
        selected_data = np.random.randint(len(self.synthetic_data), size=(len(majority_index[0]) - len(minority_index[0]), 1))
        self.synthetic_data = np.array(self.synthetic_data)[selected_data.reshape(1, -1)[0], :]

        # Taking the data to the original dimensions.
        for i in range(len(self.synthetic_data)):
            self.synthetic_data[i] = X_min + (X_max - X_min) * self.synthetic_data[i]

        # Replacing infinity values with minimal and maximal float python values.
        self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data).astype(float))

        # Creating the synthetic data frame.
        self.synthetic_df = pd.DataFrame(np.array(self.synthetic_data), columns=self.X_columns)

        # Rounding binary columns if needed.
        if self.__binarize:
            self.__to_binary()

        # Adding the target column.
        self.synthetic_df.loc[:, self.target] = minority_class
        new_df = pd.concat([self.df, self.synthetic_df], axis=0)
        return new_df


from setuptools import setup
long_description = '''
# crucio
Crucio is a python sci-kit learn inspired package for class imbalance. It use some classic methods for class balancing taking as parameters a data frame and the target column.

This version of kydavra has the next methods of feature selection:
* ADASYN.
* ICOTE (Immune Centroids Oversampling).
* MTDF (Mega-Trend Difussion Function).
* MWMOTE (Majority Weighted Minority Oversampling Technique).
* SMOTE (Synthetic Minority Oversampling Technique).
* SMOTENC (Synthetic Minority Over-sampling Technique for Nominal and Continuous).
* SMOTETOMEK (Synthetic Minority Oversampling Technique + Tomek links for undersampling).
* SMOTEENN (Synthetic Minority Oversampling Technique + ENN for undersampling).
* SCUT (SMOTE and Clustered Undersampling Technique).
* SLS (Safe-Level-Synthetic Minority Over-Sampling TEchnique).
* TKRKNN (Top-K ReverseKNN).

All these methods takes the pandas Data Frame and y column to balance on.

How to use crucio

To use balancer from crucio you should just import the balancer from crucio in the following framework:
```python
from crucio import SMOTE
```

class names are written above.\
Next create a object of this algorithm (I will use SMOTE method as an example).
```python
method = SMOTE()
```

To balance the dataset on the target column use the ‘balance’ function, using as parameters the pandas Data Frame and the column that you want to balance. Small tip, balance only the training set, not full one.

```python
new_dataframe = method.balance(df, 'target')
```

Returned value is a new data frame with the target column balanced.

With love from Sigmoid.

We are open for feedback. Please send your impression to vladimir.stojoc@gmail.com
'''

setup(
  name = 'crucio',         # How you named your package folder (MyLib)
  packages = ['crucio'],   # Chose the same as "name"
  version = '0.1.93',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Crucio is a python sci-kit learn inspired package for class imbalance. It use some classic methods for class balancing taking as parameters a data frame and the target column.',   # Give a short description about your library
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Păpăluță Vasile, Stojoc Vladimir',                   # Type in your name
  author_email = 'vladimir.stojoc@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/SigmoidAI/crucio',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/ScienceKot/crucio/archive/v1.0.tar.gz',    # I explain this later on
  keywords = ['ml', 'machine learning', 'imbalanced learning', 'class balancing', 'python'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'scikit-learn',
          'statistics'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Framework :: Jupyter',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ]
)


'''
Created with love by Sigmoid

@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
import random
import sys
from random import randrange
import math
from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn


class SLS:

    def __init__(self,k: "int > 0" = 5, seed: float = 42, binary_columns : list = None) -> None:
        '''
            Setting up the algorithm
        :param k: int, k>0, default = 5
            Number of neighbours which will be considered when looking for simmilar data points
        :param seed: intt, default = 42
            seed for random
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        '''

        self.__k = k

        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns

        self.__seed = seed
        np.random.seed(self.__seed)
        random.seed(self.__seed)

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''

        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''

        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def balance(self, df : pd.DataFrame, target : str):
        '''
            Reducing the dimensionality of the data
        :param df: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The target name of the value that we have to predict
        '''
        
        unique = df[target].unique()

        if target not in df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")

        self.df = df.copy()
        self.target = target

        #training columns
        self.X_columns = [column for column in self.df.columns if column != target]
        self.synthetic_data = []
        self.synthetic_final_df = self.df.copy()

        classes_nr_samples = []
        new_sample = []

        #check the minority and majority class
        for cl in unique:
            classes_nr_samples.append(len(df[df[target]==cl]))
            majority_class_nr_samples = max(classes_nr_samples)
        majority_class = unique[np.argmax(classes_nr_samples)]

        #minority and majority samples
        minority_classes = [cl for cl in unique if cl!=majority_class]
        minority_classes_samples = [cl for i,cl in enumerate(classes_nr_samples) if i!=np.argmax(classes_nr_samples)]

        #for every class perform oversample
        for i,minority_class in enumerate(minority_classes):

            #difference in samples between minority and majority class
            difference = majority_class_nr_samples-minority_classes_samples[i]

            #minority sample
            minority_samples = self.df[self.df[target] == minority_class][self.X_columns].values
            new_examples_nr=0
            for minority_sample in minority_samples:

                #check if already have needed number of samples
                if new_examples_nr == difference:
                    break

                #randomly select one of the neares neighbors of the sample
                minority_nearest_neighbours = self.__get_k_neighbours(minority_sample,minority_samples)
                index = randrange(len(minority_nearest_neighbours))
                neighbour_index = minority_nearest_neighbours[index]
                selected_neighbour = minority_samples[neighbour_index]

                #check the number of positive neighbour around sample an its neighbor which is safe level for itself
                positive_instances_example = self.__get_k_neighbours_positive_number(minority_sample,minority_class)
                positive_instances_neighb = self.__get_k_neighbours_positive_number(selected_neighbour,minority_class)

                #calculate the ration by dividing the save level of sample and save level of neighbor
                if positive_instances_neighb!=0:
                    ratio = positive_instances_example/positive_instances_neighb
                else :
                    ratio = -1

                #if ration == -1 do not sample a new instance
                if ratio == -1 and positive_instances_example == 0:
                    continue

                #sample new instance
                else :

                    #depending on ration, chose where to create new neigbor
                    #gap is the number that shows how close synthetic sample would be to our minority sample
                    for j in range(len(minority_sample)):
                        if ratio == -1 and positive_instances_example!=0:
                            gap = 0
                        elif ratio == 1:
                            gap = random.random()
                        elif ratio > 1:
                            gap = random.uniform(0,1/ratio)
                        elif ratio<1: 
                            gap = random.uniform(1-ratio,1)
                        dif = selected_neighbour[j]-minority_sample[j]

                        #adding new example to df
                        new_sample.append(minority_sample[j]+gap*dif)
                    self.synthetic_data.append(new_sample)
                    new_examples_nr+=1
                    new_sample = []

            self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data))
            self.synthetic_df = pd.DataFrame(np.array(self.synthetic_data), columns=self.X_columns)
            self.synthetic_data = []
            self.synthetic_df.loc[:, target] = minority_class

            self.synthetic_final_df = pd.concat([self.synthetic_df,self.synthetic_final_df],axis=0)  

        self.synthetic_df = self.synthetic_final_df.copy()
        
        # Rounding binary columns if needed.
        if self.__binarize:
            self.__to_binary()

        #return new df
        return self.synthetic_df
    



    def __get_k_neighbours(self,example,minority_samples):
        '''
            KNN, getting nearest neighbors
        :param example: Numpy array
            the sample row from minority class to get neighbours from
        :param minority_samples: Numpy.ndarray
            minority class samples from where we find neighbours
        '''

        distances = []
        for x in minority_samples:
            distances.append(np.linalg.norm(x - example, ord=2))
        predicted_index = np.argsort(distances)[1:self.__k + 1]
        
        return predicted_index

    def __get_k_neighbours_positive_number(self,example,minority_class):
        '''
            KNN, getting all positive nearest neigbors
        :param example: Numpy array
            the sample row from minority class to get neighbours from
        :param minority_samples: Numpy.ndarray
            minority class samples from where we find neighbours
        '''

        distances = []

        for x in self.df[self.X_columns].values:
            distances.append(np.linalg.norm(x - example, ord=2))
        predicted_index = np.argsort(distances)[1:self.__k + 1]
        count = 0

        for i in predicted_index:
            if self.df[self.target].values[i]==minority_class:
                count+=1
        return count


'''
Created with love by Sigmoid

@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
import random
import sys
from random import randrange
import statistics
import math
from .erorrs import NotBinaryData, NoSuchColumn

def warn(*args, **kwargs):
    pass

import warnings
warnings.warn = warn


class SMOTENC:


    def __init__(self,k: "int > 0" = 5, seed: float = 42, binary_columns : list = None) -> None:
        '''
            Setting up the algorithm
        :param k: int, k>0, default = 5
            Number of neighbours which will be considered when looking for simmilar data points
        :param seed: int, default = 42
            seed for random
        :param binary_columns: list, default = None
            The list of columns that should have binary values after balancing.
        '''

        self.__k = k

        if binary_columns is None:
            self.__binarize = False
        else:
            self.__binarize = True
            self.__binary_columns = binary_columns

        self.__seed = seed
        np.random.seed(self.__seed)
        random.seed(self.__seed)

    def __to_binary(self) -> None:
        '''
            If the :param binary_columns: is set to True then the intermediate values in binary columns will be rounded.
        '''

        for column_name in self.__binary_columns:
            serie = self.synthetic_df[column_name].values
            threshold = (self.df[column_name].max() + self.df[column_name].min()) / 2
            for i in range(len(serie)):
                if serie[i] >= threshold:
                    serie[i] = self.df[column_name].max()
                else:
                    serie[i] = self.df[column_name].min()
            self.synthetic_df[column_name] = serie

    def __infinity_check(self, matrix : 'np.array') -> 'np.array':
        '''
            This function replaces the infinity and -infinity values with the minimal and maximal float python values.
        :param matrix: 'np.array'
            The numpy array that was generated my the algorithm.
        :return: 'np.array'
            The numpy array with the infinity replaced values.
        '''

        matrix[matrix == -np.inf] = sys.float_info.min
        matrix[matrix == np.inf] = sys.float_info.max
        return matrix

    def balance(self, df : pd.DataFrame, target : str)-> pd.DataFrame:
        '''
            Balance all minority classes to the number of majority class instances
        :param df: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The target name of the value that we have to predict
        '''

        #check for binary
        unique = df[target].unique()

        if target not in df.columns:
            raise NoSuchColumn(f"{target} isn't a column of passed data frame")


        self.df = df.copy()
        self.synthetic_final_df = self.df.copy()

        #training columns
        self.X_columns = [column for column in self.df.columns if column != target]

        classes_nr_samples = []

        #Find the majority class and creating nr of minority calss samples list
        for cl in unique:
            classes_nr_samples.append(len(df[df[target]==cl]))
            majority_class_nr_samples = max(classes_nr_samples)
        majority_class = unique[np.argmax(classes_nr_samples)]

        minority_classes = [cl for cl in unique if cl!=majority_class]
        minority_classes_samples = [cl for i,cl in enumerate(classes_nr_samples) if i!=np.argmax(classes_nr_samples)]         

        #check for categorical data i 
        not_categorical_columns = [] 
        for column in self.df.columns:
            if column!=target and df[column].values.dtype!=object:
                not_categorical_columns.append(column)

        for i,minority_class in enumerate(minority_classes):

            #find difference between nr samples of minority and mojority class
            difference = majority_class_nr_samples-minority_classes_samples[i]

            #set minority class for specific class
            minority_df = self.df[self.df[target] == minority_class][self.X_columns]
            minority_samples = self.df[self.df[target] == minority_class][self.X_columns].values
            self.synthetic_data = []
            std_list = []

            #add columns for which calculate std
            for col in not_categorical_columns:
                std_list.append(np.std(minority_df[col].values))
            std_list.sort()

            #calculate mean to add for every minority class
            self.median = statistics.median(std_list)

            for _ in range(difference):

                #random example from minority class
                index = randrange(len(minority_samples))
                example = minority_samples[index]

                #select k neighbouors from this example
                neighbours_indexes = self.__get_k_neighbours(example,minority_samples) 

                #select random neighbour
                index = randrange(len(neighbours_indexes))
                selected_neighbour = neighbours_indexes[index]
                selected_neighbour = minority_samples[selected_neighbour]

                #select random point between example and neighbour
                alpha = random.random()
                new_example = []

                #depending if categorical then the most common value from neighbours, otherwise, simple SMOTE
                for feature,_ in enumerate(example):
                    if type(example[feature])!=str:
                        new_example.append(example[feature]+alpha*(selected_neighbour[feature]-example[feature]))
                    else: 
                        neighbours = pd.DataFrame(minority_samples,columns = self.X_columns )
                        feature_list = neighbours[self.X_columns[feature]].values
                        neighbours_features = []
                        for n in neighbours_indexes:
                            neighbours_features.append(feature_list[n])
                        new_example.append(self.__most_common(neighbours_features))

                #add it to df
                self.synthetic_data.append(new_example)
            self.synthetic_data = self.__infinity_check(np.array(self.synthetic_data))
            self.synthetic_df = pd.DataFrame(np.array(self.synthetic_data), columns=self.X_columns)
            
            print(self.synthetic_df.info())
            self.synthetic_data = []
            self.synthetic_df.loc[:, target] = minority_class

            #to all dataset add new rows
            self.synthetic_final_df = pd.concat([self.synthetic_df,self.synthetic_final_df],axis=0)

        
        self.synthetic_df = self.synthetic_final_df.copy()
        
        #convert back to float noncategorical columns
        for col in not_categorical_columns:
            self.synthetic_df[col] = self.synthetic_df[col].astype(float)

        # Rounding binary columns if needed.      
        if self.__binarize:
            self.__to_binary()

        #return new df
        return self.synthetic_df
    
    def __get_k_neighbours(self,example,minority_sample):
        '''
            KNN, getting nearest neighbors
        :param example: Numpy array
            the sample row from minority class to get neighbours from
        :param minority_samples: Numpy.ndarray
            minority class samples from where we find neighbours
        '''
        distances = []

        for x in minority_sample:
            dist=0
            for feature,_ in enumerate(x):
                if type(x[feature])!=str:
                    dist+=(x[feature]-example[feature])**2
                else:
                    dist+=self.median**2
            distances.append(math.sqrt(dist))

        predicted_index = np.argsort(distances)[1:self.__k + 1]
        return predicted_index

    def __most_common(self,neighbours_features):
        '''
        Getting the most common categorical value for K nearest neighbours
        '''

        word_counter = {}
        for word in neighbours_features:
            if word in word_counter:
                word_counter[word] += 1
            else:
                word_counter[word] = 1

        popular_words = sorted(word_counter, key = word_counter.get, reverse = True)

        return popular_words[0]